#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import structlog
from datetime import datetime
from transitions import Machine
from transitions.extensions.nesting import HierarchicalMachine as HMachine
from twisted.python import failure
from twisted.internet import reactor
from twisted.internet.defer import Deferred, CancelledError
from voltha_protos.voltha_pb2 import ImageDownload
from voltha_protos.omci_mib_db_pb2 import OpenOmciEventType
from pyvoltha.adapters.extensions.omci.omci_defs import ReasonCodes, OmciSectionDataSize
from pyvoltha.adapters.extensions.omci.omci_entities import SoftwareImage
from pyvoltha.adapters.extensions.omci.omci_cc import DEFAULT_OMCI_TIMEOUT
from pyvoltha.adapters.extensions.omci.omci_messages import OmciEndSoftwareDownloadResponse, OmciActivateImageResponse

###################################################################################
##              OLT out-of-band download image procedure
###################################################################################

class ImageDownloadeSTM(object):
    DEFAULT_STATES = ['disabled', 'downloading', 'validating', 'done']
    DEFAULT_TRANSITIONS = [
        {'trigger': 'start', 'source': 'disabled', 'dest': 'downloading'},
        {'trigger': 'stop',  'source': ['downloading', 'validating', 'done'], 'dest': 'disabled'},
        {'trigger': 'dw_success', 'source': 'downloading', 'dest': 'validating'},
        {'trigger': 'dw_fail', 'source': 'downloading', 'dest': 'done'},
        {'trigger': 'validate_success', 'source': 'validating', 'dest': 'done'},
    ]
    DEFAULT_TIMEOUT_RETRY = 1000      # Seconds to delay after task failure/timeout

    # def __init__(self, omci_agent, dev_id, local_name, local_dir, remote_url, download_task, 
    def __init__(self, omci_agent, image_download, 
                     download_task_cls, 
                     states=DEFAULT_STATES,
                     transitions=DEFAULT_TRANSITIONS,
                     initial_state='disabled',
                     timeout_delay=DEFAULT_TIMEOUT_RETRY,
                     advertise_events=True, clock=None):
        """
        :Param: omci_agent:	(OpenOMCIAgent)
        :Param: image_dnld: (ImageDownload)
                            ImageDownload.id  : device id
                            ImageDownload.name: file name of the image 
                            ImageDownload.url : URL to download the image from server
                            ImageDownload.local_dir: local directory of the image file
        """
        self.log = structlog.get_logger(device_id=image_download.id)
        self._agent = omci_agent
        # self._imgdw = ImageDownload()
        # self._imgdw.name = local_name
        # self._imgdw.id   = dev_id
        # self._imgdw.url  = remote_url
        # self._imgdw.local_dir = local_dir
        self._imgdw = image_download
        # self._imgdw.state = ImageDownload.DOWNLOAD_UNKNOWN   # voltha_pb2

        self._download_task_cls = download_task_cls
        self._timeout_delay   = timeout_delay

        self._current_task  = None
        self._task_deferred = None
        self._ret_deferred  = None
        self._timeout_dc = None    # DelayedCall
        self._advertise_events = advertise_events
        self.reactor = clock if clock is not None else reactor 

        self.log.debug("ImageDownloadeSTM", image_download=self._imgdw)
        self.machine = Machine(model=self, states=states,
                               transitions=transitions,
                               initial=initial_state,
                               queued=True,
                               name='{}-{}'.format(self.__class__.__name__, self._imgdw.id))
    # @property
    # def name(self):
    #     return self._imgdw.name

    def _cancel_timeout(self):
        d, self._timeout_dc = self._timeout_dc, None
        if d is not None and not d.called:
            d.cancel()

    @property
    def status(self):
        return self._imgdw

    @property
    def deferred(self):
        return self._ret_deferred
        
    def advertise(self, event, info):
        """Advertise an event on the OpenOMCI event bus"""
        if self._advertise_events:
            self._agent.advertise(event,
                                  {
                                      'state-machine': self.machine.name,
                                      'info': info,
                                      'time': str(datetime.utcnow())
                                  })

    # def reset(self):
    #     """
    #     Reset all the state machine to intial state
    #     It is used to clear failed result in last downloading
    #     """
    #     self.log.debug('reset download', image_download=self._imgdw)
    #     if self._current_task is not None:
    #         self._current_task.stop()
            
    #     self._cancel_deferred()
        
    #     if self._ret_deferred is not None:
    #         self._ret_deferred.cancel()
    #         self._ret_deferred = None

    #     self.stop()
    #     self._imgdw.state = ImageDownload.DOWNLOAD_UNKNOWN
        
    def get_file(self):
        """
          return a Deferred object
          Caller will register a callback to the Deferred to get notified once the image is available
        """
        # self.log.debug('get_file', image_download=self._imgdw)
        if self._ret_deferred is None or self._ret_deferred.called:
            self._ret_deferred = Deferred()
            
        if self._imgdw.state == ImageDownload.DOWNLOAD_SUCCEEDED:
            self.log.debug('Image Available')
            self.reactor.callLater(0, self._ret_deferred.callback, self._imgdw)
        elif self._imgdw.state == ImageDownload.DOWNLOAD_FAILED or self._imgdw.state == ImageDownload.DOWNLOAD_UNSUPPORTED:
            self.log.debug('Image not exist')
            self.reactor.callLater(0, self._ret_deferred.errback, failure.Failure(Exception('Image Download Failed ' + self._imgdw.name)))
        elif self._imgdw.state == ImageDownload.DOWNLOAD_UNKNOWN or self._imgdw.state == ImageDownload.DOWNLOAD_REQUESTED:
            self.log.debug('Start Image STM')
            self._imgdw.state = ImageDownload.DOWNLOAD_STARTED
            self.reactor.callLater(0, self.start)
        else:
            self.log.debug('NO action', state=self._imgdw.state)
            
        return self._ret_deferred
            
    def timeout(self):
        self.log.debug('Image Download Timeout', download_task=self._current_task);
        if self._current_task:
            self.reactor.callLater(0, self._current_task.stop)
        # if self._task_deferred is not None and not self._task_deferred.called:
        #     self._task_deferred.cancel()
            self._current_task = None
        # else:
        #     self.dw_fail()
            
    def on_enter_downloading(self):
        self.log.debug("on_enter_downloading")
        self.advertise(OpenOmciEventType.state_change, self.state)
        def success(results):
            self.log.debug('image-download-success', results=results)
            self._imgdw.state = ImageDownload.DOWNLOAD_SUCCEEDED
            self._imgdw.reason = ImageDownload.NO_ERROR
            self._current_task = None
            self._task_deferred = None
            self.dw_success()

        def failure(reason):
            self.log.info('image-download-failure', reason=reason)
            if self._imgdw.state == ImageDownload.DOWNLOAD_STARTED:
                self._imgdw.state = ImageDownload.DOWNLOAD_FAILED
            if isinstance(reason, CancelledError): 
                self._imgdw.reason = ImageDownload.CANCELLED
            self._current_task = None
            self._task_deferred = None
            self.dw_fail()

        self._device = self._agent.get_device(self._imgdw.id)
        self._current_task = self._download_task_cls(self._agent, self._imgdw, self.reactor)

        self._task_deferred = self._device.task_runner.queue_task(self._current_task)
        self._task_deferred.addCallbacks(success, failure)
        self._imgdw.state = ImageDownload.DOWNLOAD_STARTED

        if self._timeout_delay > 0:
            self._timeout_dc = self.reactor.callLater(self._timeout_delay, self.timeout)

    def on_enter_validating(self):
        self.log.debug("on_enter_validating")
        self.advertise(OpenOmciEventType.state_change, self.state)
        self.validate_success()

    def on_enter_done(self):
        self.log.debug("on_enter_done")
        self.advertise(OpenOmciEventType.state_change, self.state)
        self._cancel_timeout()

        d, self._ret_deferred = self._ret_deferred, None
        if d is not None:
            if self._imgdw.state == ImageDownload.DOWNLOAD_SUCCEEDED:
                self.reactor.callLater(0, d.callback, self._imgdw)
            else:  # failed
                if self._imgdw.reason == ImageDownload.CANCELLED:
                    self.reactor.callLater(0, d.cancel)
                else:
                    self.reactor.callLater(0, d.errback, failure.Failure(Exception('Image Download Failed ' + self._imgdw.name)))
            
    def on_enter_disabled(self):
        self.log.debug("on_enter_disabled")
        self.advertise(OpenOmciEventType.state_change, self.state)

        self._cancel_timeout()
        if self._current_task is not None:
            self.reactor.callLater(0, self._current_task.stop)
            self._current_task = None

        if self._ret_deferred:
            self.reactor.callLater(0, self._ret_deferred.cancel)
            self._ret_deferred = None
            
        # remove local file fragments if download failed
        file_path = self._imgdw.local_dir + '/' + self._imgdw.name
        if self._imgdw.state != ImageDownload.DOWNLOAD_SUCCEEDED and os.path.exists(file_path):
            os.remove(file_path)            
        self._imgdw.state = ImageDownload.DOWNLOAD_UNKNOWN

###################################################################################
##              OMCI Software Image Download Procedure
###################################################################################

class OmciSoftwareImageDownloadSTM(object):
    
    OMCI_SWIMG_DOWNLOAD_TIMEOUT = 5400      # TODO: Seconds for the full downloading procedure to avoid errors that cause infinte downloading
    OMCI_SWIMG_DOWNLOAD_WINDOW_SIZE = 32
    OMCI_SWIMG_WINDOW_RETRY_MAX = 2
    OMCI_SWIMG_ACTIVATE_RETRY_MAX = 2
    OMCI_SWIMG_ACTIVATE_TRANSITIONS_TIMEOUT = 10      # Seconds to delay after task failure/timeout

    # def __init__(self, omci_agent, dev_id, img_path, 
    def __init__(self, image_id, omci_agent, image_dnld,
                     window_size=OMCI_SWIMG_DOWNLOAD_WINDOW_SIZE,
                     timeout_delay=OMCI_SWIMG_DOWNLOAD_TIMEOUT,
                     advertise_events=True,
                     clock=None):
        """
        omci_agent:	(OpenOMCIAgent)
        image_dnld: (ImageDownload)
                    ImageDownload.id  : device id
                    ImageDownload.name: file name of the image 
                    ImageDownload.url : URL to download the image from server
                    ImageDownload.local_dir: local directory of the image file
        window_size: window size of OMCI download procedure 
        """
        self.log = structlog.get_logger(device_id=image_dnld.id)
        self._omci_agent = omci_agent
        self._image_download = image_dnld
        self._timeout = timeout_delay
        self._timeout_dc = None
        self._window_size = window_size
        self.reactor = clock if clock is not None else reactor
        self._offset = 0
        # self._win_section = 0
        self._win_retry = 0
        self._device_id = image_dnld.id
        self._device = omci_agent.get_device(image_dnld.id)
        self.__init_state_machine()
        self._ret_deferred = None
        self._image_id = image_id    # Target software image entity ID
        self._image_file = image_dnld.local_dir + '/' + image_dnld.name
        self._image_obj = open(self._image_file, mode='rb')
        self._image_size = os.path.getsize(self._image_file)
        self._crc32 = 0
        self._win_crc32 = 0
        self._win_data = None
        self._current_deferred = None
        self._result = None    # ReasonCodes
        self.crctable = []
        self._crctable_init = False
        self._actimg_retry_max = OmciSoftwareImageDownloadSTM.OMCI_SWIMG_ACTIVATE_RETRY_MAX
        self._actimg_retry = 0
        self.log.debug("DownloadSTM", image=self._image_file, image_size=self._image_size)
 
    def __init_state_machine(self):

        #### Download Window Sub State Machine ####
        OMCI_DOWNLOAD_WINDOW_STATE = ['init_window', 'sending_sections', 'window_success', 'window_failed']
        OMCI_DOWNLOAD_WINDOW_TRANSITIONS = [
            {'trigger': 'send_sections',  'source': 'init_window',     'dest': 'sending_sections'},
            # {'trigger': 'send_section_last',  'source': 'start_section', 'dest': 'last_section'  },
            {'trigger': 'rx_ack_success', 'source': 'sending_sections', 'dest': 'window_success' },
            {'trigger': 'rx_ack_failed',  'source': 'sending_sections', 'dest': 'window_failed'  },
            # {'trigger': 'retry_window',   'source': 'window_failed', 'dest': 'start_section'  },
            {'trigger': 'reset_window',   'source': '*',               'dest': 'init_window'    }
        ]    
        self.win_machine = HMachine(model=self, 
                                    states=OMCI_DOWNLOAD_WINDOW_STATE,
                                    transitions=OMCI_DOWNLOAD_WINDOW_TRANSITIONS,
                                    initial='init_window',
                                    queued=True,
                                    name='{}-window_section_machine'.format(self.__class__.__name__))

        #### Software Activation Sub State Machine ####
        OMCI_SWIMG_ACTIVATE_STATES = ['init_act', 'activating', 'busy', 'rebooting', 'committing', 'done', 'failed']
        OMCI_SWIMG_ACTIVATE_TRANSITIONS = [
            {'trigger': 'activate', 'source': ['init_act', 'busy'], 'dest': 'activating'},
            {'trigger': 'onu_busy', 'source': 'activating', 'dest': 'busy'},
            {'trigger': 'reboot',   'source': 'activating', 'dest': 'rebooting'},
            {'trigger': 'do_commit', 'source': ['activating', 'rebooting'], 'dest': 'committing'},
            # {'trigger': 'commit_ok', 'source': 'committing', 'dest': 'done'},
            {'trigger': 'reset_actimg', 'source': ['activating', 'rebooting', 'committing', 'failed'], 'dest': 'init_act'},
            # {'trigger': 'actimg_fail', 'source': ['init_act', 'activating', 'rebooting', 'committing'], 'dest': 'failed'}
        ]
        
        self.activate_machine = HMachine(model=self, 
                                         states=OMCI_SWIMG_ACTIVATE_STATES,
                                         transitions=OMCI_SWIMG_ACTIVATE_TRANSITIONS,
                                         initial='init_act',
                                         queued=True,
                                         name='{}-activate_machine'.format(self.__class__.__name__))
                                   
        #### Main State Machine ####
        OMCI_SWIMG_DOWNLOAD_STATES = [ 'init_image', 'starting_image', 'ending_image', 'endimg_busy', 'done_image',
                                      {'name': 'dwin', 'children': self.win_machine}, 
                                      {'name': 'actimg', 'children': self.activate_machine}
                                     ]
        OMCI_SWIMG_DOWNLOAD_TRANSITIONS = [
            {'trigger': 'start_image',      'source': 'init_image',     'dest': 'starting_image' },
            {'trigger': 'download_window',  'source': 'starting_image', 'dest': 'dwin_init_window' },
            {'trigger': 'download_success', 'source': 'dwin',           'dest': 'ending_image'   },
            {'trigger': 'onu_busy',         'source': 'ending_image',   'dest': 'endimg_busy'    },
            {'trigger': 'retry_endimg',     'source': 'endimg_busy',    'dest': 'ending_image'   },
            {'trigger': 'end_img_success',  'source': 'ending_image',   'dest': 'actimg_init_act'  },
            {'trigger': 'activate_done',    'source': 'actimg',         'dest': 'done_image'     },
            {'trigger': 'download_fail',    'source': '*',              'dest': 'done_image'     },
            {'trigger': 'reset_image',      'source': '*',              'dest': 'init_image'     },
        ]
        
        self.img_machine = HMachine(model=self, 
                                   states=OMCI_SWIMG_DOWNLOAD_STATES,
                                   transitions=OMCI_SWIMG_DOWNLOAD_TRANSITIONS,
                                   initial='init_image',
                                   queued=True,
                                   name='{}-image_download_machine'.format(self.__class__.__name__))

    # @property
    # def image_filename(self):
    #     return self._image_file

    # @image_filename.setter
    # def image_filename(self, value):
    #     if self._image_fd is not None:
    #         self._image_fd.close()
    #     self._image_filename = value
    #     self._image_fd = open(self._image_filename, mode='rb')
    #     self._image_size = os.path.getsize(self._image_filename)
    #    print("Set image file: " + self._image_filename + " size: " + str(self._image_size))

    def __omci_start_download_resp_success(self, rx_frame):
        self.log.debug("__omci_download_resp_success")
        self.download_window()
        return rx_frame

    def __omci_start_download_resp_fail(self, fail):
        self.log.debug("__omci_download_resp_fail", failure=fail)
        self._result = ReasonCodes.ProcessingError
        self.download_fail()

    def __omci_end_download_resp_success(self, rx_frame):
        self.log.debug("__omci_end_download_resp_success")
        if rx_frame.fields['message_type'] == OmciEndSoftwareDownloadResponse.message_id: # 0x35
            omci_data = rx_frame.fields['omci_message']
            if omci_data.fields['result'] == 0: 
                self.log.debug('OMCI End Image OK')
                self._result = ReasonCodes.Success
                self.end_img_success()
            elif omci_data.fields['result'] == 6: # Device Busy
                self.log.debug('OMCI End Image Busy')
                self.onu_busy()
            else:
                self.log.debug('OMCI End Image Failed', reason=omci_data.fields['result'])
        else:
            self.log.debug('Receive Unexpected OMCI', message_type=rx_frame.fields['message_type'])

    def __omci_end_download_resp_fail(self, fail):
        self.log.debug("__omci_end_download_resp_fail", failure=fail)
        self._result = ReasonCodes.ProcessingError
        self.download_fail()
    
    def __omci_send_window_resp_success(self, rx_frame, cur_state, datasize):
        # self.log.debug("__omci_send_window_resp_success", current_state=cur_state)
        self._offset += datasize
        self._image_download.downloaded_bytes += datasize
        self.rx_ack_success()

    def __omci_send_window_resp_fail(self, fail, cur_state):
        self.log.debug("__omci_send_window_resp_fail", current_state=cur_state)
        self.rx_ack_failed()

    def __activate_resp_success(self, rx_frame):
        self._current_deferred = None
        if rx_frame.fields['message_type'] == OmciActivateImageResponse.message_id: # 0x36
            omci_data = rx_frame.fields['omci_message']
            if omci_data.fields['result'] == 0:
                self.log.debug("Activate software image success, rebooting ONU ...", device_id=self._device.device_id,
                                state=self._image_download.image_state)
                standby_image_id = 0 if self._image_id else 1
                self._omci_agent.database.set(self._device.device_id, SoftwareImage.class_id, self._image_id, 	{"is_active": 1})
                self._omci_agent.database.set(self._device.device_id, SoftwareImage.class_id, standby_image_id, {"is_active": 0})
                self.reboot()
            elif omci_data.fields['result'] == 6: # Device Busy
                self.log.debug('OMCI Activate Image Busy')
                self.onu_busy()
            else:
                self.log.debug('OMCI Activate Image Failed', reason=omci_data['result'])
        else:
            self.log.debug('Receive Unexpected OMCI', message_type=rx_frame['message_type'])
                
    def __activate_fail(self, fail):
        self.log.debug("Activate software image failed", faile=fail)
        self._current_deferred = None
        self._result = ReasonCodes.ProcessingError
        self.activate_done()
        
    def __commit_success(self, rx_frame):
        self.log.debug("Commit software success", device_id=self._device_id)
        self._current_deferred = None
        standby_image_id = 0 if self._image_id else 1
        self._omci_agent.database.set(self._device_id, SoftwareImage.class_id, self._image_id, {"is_committed": 1})
        self._omci_agent.database.set(self._device_id, SoftwareImage.class_id, standby_image_id, {"is_committed": 0})
        self._image_download.image_state = ImageDownload.IMAGE_ACTIVE
        self._result = ReasonCodes.Success
        self.activate_done()

    def __commit_fail(self, fail):
        self.log.debug("Commit software image failed", faile=fail)
        self._current_deferred = None
        self._result = ReasonCodes.ProcessingError
        self._image_download.image_state = ImageDownload.IMAGE_REVERT
        self.activate_done()

#    @property
#    def image_id(self):
#        return self._image_id

#    @image_id.setter
#    def image_id(self, value):
#        self._image_id = value

    @property
    def status(self):
        return self._image_download

    def start(self):
        self.log.debug("OmciSoftwareImageDownloadSTM.start", current_state=self.state)
        if self._ret_deferred is None:
            self._ret_deferred = Deferred()
        if self.state == 'init_image':
            self.reactor.callLater(0, self.start_image)
        return self._ret_deferred

    def stop(self):
        self.log.debug("OmciSoftwareImageDownloadSTM.stop", current_state=self.state)
        self._result = ReasonCodes.OperationCancelled
        self.download_fail()
        
    def on_enter_init_image(self):
        self.log.debug("on_enter_init_image")
        self._image_obj.seek(0)
        self._offset = 0
        # self._win_section = 0
        self._win_retry   = 0
        
    def on_enter_starting_image(self):
        self.log.debug("on_enter_starting_image")
        self._image_download.downloaded_bytes = 0
        self._current_deferred = self._device.omci_cc.send_start_software_download(self._image_id, self._image_size, self._window_size)
        self._current_deferred.addCallbacks(self.__omci_start_download_resp_success, self.__omci_start_download_resp_fail)
                                            # callbackArgs=(self.state,), errbackArgs=(self.state,))

    def on_enter_dwin_init_window(self):
        # self.log.debug("on_enter_dwin_init_window", offset=self._offset, image_size=self._image_size)
        if self._offset < self._image_size:
            self.send_sections()

    def on_enter_dwin_sending_sections(self):
        # self.log.debug("on_enter_dwin_sending_sections", offset=self._offset)

        if (self._offset + self._window_size * OmciSectionDataSize) <= self._image_size:
            sections = self._window_size
            mod = 0
            datasize = self._window_size * OmciSectionDataSize
        else:
            datasize = self._image_size - self._offset
            sections = datasize / OmciSectionDataSize
            mod = datasize % OmciSectionDataSize
            sections = sections + 1 if mod > 0 else sections

        # self.log.debug("on_enter_dwin_sending_sections", offset=self._offset, datasize=datasize, sections=sections)
        if self._win_retry == 0:
            self._win_data = self._image_obj.read(datasize)
            self._win_crc32 = self.crc32(self._crc32, self._win_data)
            # self.log.debug("CRC32", crc32=self._win_crc32, offset=self._offset)
        else:
            self.log.debug("Retry download window with crc32", offset=self._offset)
            
        sent = 0
        for i in range(0, sections):
            if i < sections - 1:
                # self.log.debug("section data", data=hexlify(data[(self._offset+sent):(self._offset+sent+OmciSectionDataSize)]))
                self._device.omci_cc.send_download_section(self._image_id, i,
                                                           self._win_data[sent:sent+OmciSectionDataSize])
                sent += OmciSectionDataSize
            else:
                last_size = OmciSectionDataSize if mod == 0 else mod
                self._current_deferred = self._device.omci_cc.send_download_section(self._image_id, i,
                                                           self._win_data[sent:sent+last_size],
                                                           timeout=DEFAULT_OMCI_TIMEOUT)
                self._current_deferred.addCallbacks(self.__omci_send_window_resp_success, self.__omci_send_window_resp_fail,
                                                    callbackArgs=(self.state, datasize), errbackArgs=(self.state,))
                sent += last_size
                assert sent==datasize

    # def on_enter_dwin_last_section(self):
    #     self._current_deferred = self._device.omci_cc.send_download_section, self._instance_id, self._win_section, data)
    #     self._current_deferred.addCallbacks(self.__omci_resp_success, self.__omci_resp_fail,
    #                                         callbackArgs=(self.state,), errbackArgs=(self.state,))

    def on_enter_dwin_window_success(self):
        # self.log.debug("on_enter_dwin_window_success")
        self._crc32 = self._win_crc32 if self._win_crc32 != 0 else self._crc32
        self._win_crc32 = 0
        self._win_retry = 0
        if self._offset < self._image_size:
            self.reset_window()
        else:
            self.download_success()

    def on_enter_dwin_window_failed(self):
        self.log.debug("on_enter_dwin_window_fail: ", retry=self._win_retry)
        if self._win_retry < self.OMCI_SWIMG_WINDOW_RETRY_MAX:
            self._win_retry += 1
            self.reset_window()
        else:
            self._result = ReasonCodes.ProcessingError
            self.download_fail()

    def on_enter_ending_image(self):
        self.log.debug("on_enter_ending_image", crc32=self._crc32)
        self._current_deferred = self._device.omci_cc.send_end_software_download(self._image_id, self._crc32, 
                                                                                 self._image_size, timeout=18)
        self._current_deferred.addCallbacks(self.__omci_end_download_resp_success, self.__omci_end_download_resp_fail)
                                            # callbackArgs=(self.state,), errbackArgs=(self.state,))

    def on_enter_endimg_busy(self):
        self.log.debug("on_enter_endimg_busy")
        self.reactor.callLater(3, self.retry_endimg)

    def on_enter_actimg_init_act(self):
        self.log.debug("on_enter_actimg_init_act", retry=self._actimg_retry, max_retry=self._actimg_retry_max)
        # self._images[0] = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 0, ["is_active", "is_committed", "is_valid"])
        # self._images[1] = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 1, ["is_active", "is_committed", "is_valid"])
        # if (self._images[self._to_image]["is_active"] != 1 and self._images[self._to_image]["is_valid"] == 1):
        if self._actimg_retry > self._actimg_retry_max:
            self.log.debug("activate image failed: retry max", retries=self._actimg_retry)
            self._result = ReasonCodes.ProcessingError
            self.activate_done()
        else:
            self._image_download.image_state = ImageDownload.IMAGE_ACTIVATE
            self.activate()
            
    def on_enter_actimg_activating(self):
        self.log.debug("on_enter_actimg_activating")
        img = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 
                                              self._image_id, ["is_active", "is_committed", "is_valid"])
                                              
        self.log.debug("on_enter_actimg_activating", instance=self._image_id, state=img)
        if img["is_active"] == 0:
            #if img["is_valid"] == 1:
            self._current_deferred = self._device.omci_cc.send_active_image(self._image_id)
            self._current_deferred.addCallbacks(self.__activate_resp_success, self.__activate_fail)
            #else:
            #    self.fail()
        else:
            self.do_commit()

    def on_enter_actimg_busy(self):
        self.log.debug("on_enter_actimg_busy")
        self.reactor.callLater(3, self.activate)
        
    def __on_reboot_timeout(self):
        self.log.debug("on_reboot_timeout")
        self._timeout_dc = None
        self._result = ReasonCodes.ProcessingError
        self.activate_done()
        
    def on_enter_actimg_rebooting(self):
        self.log.debug("on_enter_actimg_rebooting")
        if self._timeout_dc == None:
            self._timeout_dc = self.reactor.callLater(self._timeout, self.__on_reboot_timeout)

    def on_exit_actimg_rebooting(self):
        self.log.debug("on_exit_actimg_rebooting", timeout=self._timeout_dc)
        if self._timeout_dc and self._timeout_dc.active:
            self._timeout_dc.cancel()
            self._timeout_dc = None
    
    def on_enter_actimg_committing(self):
        # self.log.debug("on_enter_committing")
        img = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 
                                              self._image_id, ["is_active", "is_committed", "is_valid"])
        self.log.debug("on_enter_actimg_committing", instance=self._image_id, state=img)
        if (img['is_active'] == 0):
            self._actimg_retry += 1
            self.log.debug("do retry", retry=self._actimg_retry)
            self.reset_actimg()
        else:
            self._actimg_retry = 0
            self._current_deferred = self._device.omci_cc.send_commit_image(self._image_id)
            self._current_deferred.addCallbacks(self.__commit_success, self.__commit_fail)

    def on_enter_done_image(self):
        self.log.debug("on_enter_done_image", result=self._result)
        if self._result == ReasonCodes.Success:
            self.reactor.callLater(0, self._ret_deferred.callback, self._image_download) # (str(self._instance_id))
        else:
            self._ret_deferred.errback(failure.Failure(Exception('ONU Software Download Failed, instance ' + str(self._image_id))))

    def __crc_GenTable32(self):
        if self._crctable_init:
            return
            
        #  x32 + x26 + x23 + x22 + x16 + x12 + x11 + x10 + x8 + x7 + x5 + x4 + x2 + x + 1   
        pn32 = [0, 1, 2, 4, 5, 7, 8, 10, 11, 12, 16, 22, 23, 26]
        poly = 0
        for i in pn32:
            poly |= (1 << i)

        for i in range(0, 256):
            _accum = (i << 24) & 0xFFFFFFFF
            for j in range(0, 8):
                if _accum & (1 << 31):
                    _accum = (_accum << 1) ^ poly
                else:
                    _accum = (_accum << 1) & 0xFFFFFFFF
            # self.crctable[i] = accum
            self.crctable.append(_accum)
        self._crctable_init = True
            
    def crc32(self, accum, data):
        self.__crc_GenTable32()
        _accum = ~accum & 0xFFFFFFFF
        num = len(data)
        for i in range(0, num):
            _accum = self.crctable[((_accum >> 24) ^ ord(data[i])) & 0xFF] ^ ((_accum << 8) & 0xFFFFFFFF)

        return ~_accum & 0xFFFFFFFF

###################################################################################
##              OMCI Software Image Activation/Committing Procedure
###################################################################################
'''
class OmciSoftwareImageActivateSTM(object):
    OMCI_SWIMG_ACTIVATE_STATES = ['starting', 'activating', 'busy', 'rebooting', 'committing', 'done', 'failed']
    OMCI_SWIMG_ACTIVATE_TRANSITIONS = [
        {'trigger': 'activate', 'source': ['starting', 'busy'], 'dest': 'activating'},
        {'trigger': 'onu_busy', 'source': 'activating', 'dest': 'busy'},
        {'trigger': 'reboot',   'source': 'activating', 'dest': 'rebooting'},
        {'trigger': 'do_commit', 'source': ['activating', 'rebooting'], 'dest': 'committing'},
        {'trigger': 'commit_ok', 'source': 'committing', 'dest': 'done'},
        {'trigger': 'reset',    'source': ['activating', 'rebooting', 'committing', 'failed'], 'dest': 'starting'},
        {'trigger': 'fail',     'source': ['starting', 'activating', 'rebooting', 'committing'], 'dest': 'failed'}
    ]
    OMCI_SWIMG_ACTIVATE_TRANSITIONS_TIMEOUT = 10      # Seconds to delay after task failure/timeout
    OMCI_SWIMG_ACTIVATE_RETRY_MAX           = 2
    def __init__(self, omci_agent, dev_id, target_img_entity_id, image_download,
                     states=OMCI_SWIMG_ACTIVATE_STATES,
                     transitions=OMCI_SWIMG_ACTIVATE_TRANSITIONS,
                     initial_state='disabled',
                     timeout_delay=OMCI_SWIMG_ACTIVATE_TRANSITIONS_TIMEOUT,
                     advertise_events=True,
                     clock=None):
        self.log = structlog.get_logger(device_id=dev_id)
        self._omci_agent = omci_agent
        self._device_id  = dev_id
        self._device = omci_agent.get_device(dev_id)
        self._to_image = target_img_entity_id
        self._from_image = 0 if self._to_image == 1 else 1
        self._image_download = image_download
        # self._images = dict()
        self._timeout = timeout_delay
        self._timeout_dc = None
        self.reactor = clock if clock is not None else reactor
        self._retry_max = OmciSoftwareImageActivateSTM.OMCI_SWIMG_ACTIVATE_RETRY_MAX
        self._retry = 0
        self._deferred = None
        self.ret_deferred = None
        self.machine = Machine(model=self, 
                               states=states,
                               transitions=transitions,
                               initial='starting',
                               queued=True,
                               name='{}-image_activate_machine'.format(self.__class__.__name__))
        self.log.debug("OmciSoftwareImageActivateSTM", target=self._to_image)

    def __activate_resp_success(self, rx_frame):
        if rx_frame.fields['message_type'] == 0x36:  # (OmciActivateImageResponse)
            omci_data = rx_frame.fields['omci_message']
            if omci_data.fields['result'] == 0:
                self.log.debug("Activate software image success, rebooting ONU ...", device_id=self._device_id) 
                self._omci_agent.database.set(self._device_id, SoftwareImage.class_id, self._to_image, 	{"is_active": 1})
                self._omci_agent.database.set(self._device_id, SoftwareImage.class_id, self._from_image, {"is_active": 0})
                self.reboot()
            elif omci_data.fields['result'] == 6: # Device Busy
                self.log.debug('OMCI Activate Image Busy')
                self.onu_busy()
            else:
                self.log.debug('OMCI Activate Image Failed', reason=omci_data['result'])
        else:
            self.log.debug('Receive Unexpected OMCI', message_type=rx_frame['message_type'])
                
    def __activate_fail(self, fail):
        self.log.debug("Activate software image failed", faile=fail)
        
    def __commit_success(self, rx_frame):
        self.log.debug("Commit software success", device_id=self._device_id)
        self._omci_agent.database.set(self._device_id, SoftwareImage.class_id, self._to_image, {"is_committed": 1})
        self._omci_agent.database.set(self._device_id, SoftwareImage.class_id, self._from_image, {"is_committed": 0})
        self.commit_ok()

    def __commit_fail(self, fail):
        self.log.debug("Commit software image failed", faile=fail)

    @property
    def status(self):
        return self._image_download
        
    def start(self):
        self.log.debug("Start switch software image", target=self._to_image)
        # self._images[0] = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 0, ["is_active", "is_committed", "is_valid"])
        # self._images[1] = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 1, ["is_active", "is_committed", "is_valid"])
        # if (self._images[self._to_image]["is_active"] == 0 and self._images[self._to_image]["is_valid"] == 1):
        self.ret_deferred = Deferred()
        self._image_download.image_state = ImageDownload.IMAGE_ACTIVATE
        self.reactor.callLater(0, self.activate)
        return self.ret_deferred

    def on_enter_starting(self):
        # self.log.debug("on_enter_starting")
        # self._images[0] = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 0, ["is_active", "is_committed", "is_valid"])
        # self._images[1] = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 1, ["is_active", "is_committed", "is_valid"])
        # if (self._images[self._to_image]["is_active"] != 1 and self._images[self._to_image]["is_valid"] == 1):
        if self._retry > self._retry_max:
            self.log.debug("failed: retry max", retries=self._retry)
            self.fail()
        else:
            self.activate()
            
    def on_enter_activating(self):
        img = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 
                                              self._to_image, ["is_active", "is_committed", "is_valid"])
                                              
        self.log.debug("on_enter_activating", instance=self._to_image, state=img)
        if img["is_active"] == 0:
            #if img["is_valid"] == 1:
            self._deferred = self._device.omci_cc.send_active_image(self._to_image)
            self._deferred.addCallbacks(self.__activate_resp_success, self.__activate_fail)
            #else:
            #    self.fail()
        else:
            self.do_commit()

    def on_enter_busy(self):
        self.log.debug("on_enter_busy")
        self.reactor.callLater(3, self.activate)
        
    def on_enter_rebooting(self):
        self.log.debug("on_enter_rebooting")
        if self._timeout_dc == None:
            self._timeout_dc = self.reactor.callLater(self._timeout, self.fail)

    def on_exit_rebooting(self):
        self.log.debug("on_exit_rebooting")
        if self._timeout_dc and self._timeout_dc.active:
            self._timeout_dc.cancel()
            self._timeout_dc = None
    
    def on_enter_committing(self):
        # self.log.debug("on_enter_committing")
        img = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 
                                              self._to_image, ["is_active", "is_committed", "is_valid"])
        self.log.debug("on_enter_committing", instance=self._to_image, state=img)
        if (img['is_active'] == 0):
            self._retry += 1
            self.log.debug("do retry", retry=self._retry)
            self.reset()
        else:
            self._retry = 0
            self._deferred = self._device.omci_cc.send_commit_image(self._to_image)
            self._deferred.addCallbacks(self.__commit_success, self.__commit_fail)

    def on_enter_done(self):
        self.log.debug("on_enter_done")
        self._image_download.image_state = ImageDownload.IMAGE_ACTIVE
        self.ret_deferred.callback(self._to_image)

    def on_enter_failed(self):
        self.log.debug("on_enter_failed")
        self._image_download.image_state = ImageDownload.IMAGE_REVERT
        self.ret_deferred.errback(failure.Failure(Exception('ONU Software Activating Failed, instance ' + str(self._to_image))))
'''

###################################################################################
##              Image Agent for OLT/ONT software image handling
###################################################################################
class ImageAgent(object):
    """
        Image Agent supports multiple state machines running at the same time:
    """

    DEFAULT_LOCAL_ROOT = "/"
    
    # def __init__(self, omci_agent, dev_id, stm_cls, img_tasks, advertise_events=True):
    def __init__(self, omci_agent, dev_id, 
                     dwld_stm_cls, dwld_img_tasks, 
                     upgrade_onu_stm_cls, upgrade_onu_tasks, 
                     # image_activate_stm_cls, 
                     advertise_events=True, local_dir=None, clock=None):
        """
        Class initialization

        :param omci_agent: (OpenOmciAgent) Agent
        :param dev_id    : (str) ONU Device ID
        :param dwld_stm_cls          : (ImageDownloadeSTM) Image download state machine class
        :param dwld_img_tasks        : (FileDownloadTask) file download task
        :param upgrade_onu_stm_cls   : (OmciSoftwareImageDownloadSTM) ONU Image upgrade state machine class
        :param upgrade_onu_tasks     : ({OmciSwImageUpgradeTask})
        # :param image_activate_stm_cls: (OmciSoftwareImageActivateSTM)
        """
        
        self.log = structlog.get_logger(device_id=dev_id)

        self._omci_agent = omci_agent
        self._device_id = dev_id
        self._dwld_stm_cls  = dwld_stm_cls
        # self._image_download_sm = None
        self._images = dict()
        self._download_task_cls = dwld_img_tasks['download-file'] 

        self._omci_upgrade_sm_cls = upgrade_onu_stm_cls
        self._omci_upgrade_task_cls = upgrade_onu_tasks['omci_upgrade_task']
        self._omci_upgrade_task = None
        self._omci_upgrade_deferred = None
        
        # self._omci_activate_img_sm_cls = image_activate_stm_cls
        # self._omci_activate_img_sm = None
        self.reactor = clock if clock is not None else reactor

        self._advertise_events = advertise_events
        # self._local_dir = None

        self._device = None
        # onu_dev = self._omci_agent.get_device(self._device_id)
        # assert device
        
        # self._local_dir = DEFAULT_LOCAL_ROOT + onu_dev.core_proxy.name
        # self.log.debug("ImageAgent", local_dir=self._local_dir)
        
        
    def __get_standby_image_instance(self):
        instance_id = None
        instance_0 = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 0, ["is_active", "is_committed"])
        if instance_0['is_active'] == 1:
            instance_id = 1
        else:
            instance_1 = self._omci_agent.database.query(self._device_id, SoftwareImage.class_id, 1, ["is_active", "is_committed"])
            if instance_1['is_active'] == 1:
                instance_id = 0
        return instance_id

    def __clear_task(self, arg):
        self.__omci_upgrade_task = None

    # def get_image(self, name, local_dir, remote_url, timeout_delay=ImageDownloadeSTM.DEFAULT_TIMEOUT_RETRY):
    def get_image(self, image_download, timeout_delay=ImageDownloadeSTM.DEFAULT_TIMEOUT_RETRY):

        """
         Get named image from servers
        :param image_download: (voltha_pb2.ImageDownload)
        :param timeout_delay : (number) timeout for download task
        :
        :Return a Deferred that will be triggered if the file is locally availabe or downloaded sucessfully
        :  Caller will register callback and errback to the returned defer to get notified
        """
        self.log.debug("get_image", download=image_download)

        # if self._local_dir is None:
        #     onu_dev = self._omci_agent.get_device(self._device_id)
        #     assert onu_dev
        #     if image_download.local_dir is None:
        #         self._local_dir = ImageAgent.DEFAULT_LOCAL_ROOT + onu_dev.core_proxy.name
        #     else:
        #         self._local_dir = image_download.local_dir + '/' + onu_dev.core_proxy.name
            
            # self.log.debug("ImageAgent", local_dir=self._local_dir)
        #     image_download.local_dir = self._local_dir
            
        # if os.path.isfile(self._local_dir + '/' + image_download.name): # image file exists
        #     d = Deferred()
        #     self.reactor.callLater(0, d.callback, image_download)
        #     self.log.debug("Image file exists")
        #     return d

        img_dnld_sm = self._images.get(image_download.name)
        if img_dnld_sm is None:
            img_dnld_sm = self._dwld_stm_cls(self._omci_agent, # self._device_id, name, local_dir, remote_url, 
                                             image_download,
                                             self._download_task_cls,
                                             timeout_delay=timeout_delay,
                                             clock=self.reactor
                                            )
            self._images[image_download.name] = img_dnld_sm

        # if self._image_download_sm is None:
        #     self._image_download_sm = self._dwld_stm_cls(self._omci_agent, # self._device_id, name, local_dir, remote_url, 
        #                                                  image_download,
        #                                                  self._download_task_cls,
        #                                                  timeout_delay=timeout_delay,
        #                                                  clock=self.reactor
        #                                                 )
        # else:
        #     if self._image_download_sm.download_status.state != ImageDownload.DOWNLOAD_SUCCEEDED:
        #         self._image_download_sm.reset()
            
        d = img_dnld_sm.get_file()
        return d

    def cancel_download_image(self, name):
        img_dnld_sm = self._images.pop(name, None)
        if img_dnld_sm is not None:
            img_dnld_sm.stop()
            
            
    def onu_omci_download(self, image_dnld_name):
        """
        Start upgrading ONU.
        image_dnld: (ImageDownload)
        : Return Defer instance to get called after upgrading success or failed. 
        : Or return None if image does not exist
        """
        self.log.debug("onu_omci_download", image=image_dnld_name)

        image_dnld_sm = self._images.get(image_dnld_name)
        if image_dnld_sm is None:
            return None
            
        self._device = self._omci_agent.get_device(image_dnld_sm.status.id) if self._device is None else self._device

        # if restart:
        #     self.cancel_upgrade_onu()            
            
        if self._omci_upgrade_task is None:
            img_id = self.__get_standby_image_instance()
            self.log.debug("start task", image_Id=img_id, task=self._omci_upgrade_sm_cls)
            self._omci_upgrade_task = self._omci_upgrade_task_cls(img_id, 
                                                                  self._omci_upgrade_sm_cls, 
                                                                  self._omci_agent, 
                                                                  image_dnld_sm.status, clock=self.reactor)
            self.log.debug("task created but not started")
            # self._device.task_runner.start()
            self._omci_upgrade_deferred = self._device.task_runner.queue_task(self._omci_upgrade_task)
            self._omci_upgrade_deferred.addBoth(self.__clear_task)
        return self._omci_upgrade_deferred


    def cancel_upgrade_onu(self):
        self.log.debug("cancel_upgrade_onu")
        if self._omci_upgrade_task is not None:
            self.log.debug("cancel_upgrade_onu", running=self._omci_upgrade_task.running)
            # if self._omci_upgrade_task.running:
            self._omci_upgrade_task.stop()
            self._omci_upgrade_task = None
        if self._omci_upgrade_deferred is not None:
           self.reactor.callLater(0, self._omci_upgrade_deferred.cancel)
           self._omci_upgrade_deferred = None
           

    # def activate_onu_image(self, image_name):
    #     self.log.debug("activate_onu_image", image=image_name)
    #     img_dnld = self.get_image_status(image_name)
    #     if img_dnld is None:
    #         return None
            
    #     img_dnld.image_state = ImageDownload.IMAGE_INACTIVE    
    #     if self._omci_activate_img_sm is None:
    #         self._omci_activate_img_sm = self._omci_activate_img_sm_cls(self._omci_agent, self._device_id,
    #                                                                     self.__get_standby_image_instance(), 
    #                                                                     img_dnld, clock=self.reactor)
    #         return self._omci_activate_img_sm.start()
    #     else:
    #         return None
            
    def onu_bootup(self):
        if self._omci_upgrade_task is not None:
            self._omci_upgrade_task.onu_bootup()

    def get_image_status(self, image_name):
        """
          Return (ImageDownload)
        """
        sm = self._images.get(image_name)
        return sm.status if sm is not None else None
    



# Copyright 2017-present Open Networking Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from unittest import TestCase, main

from pyvoltha.common.utils.ordered_weakvalue_dict import OrderedWeakValueDict


class O(object):
    def __init__(self, a):
        self.a = a


class TestOrderedWeakValueDict(TestCase):

    def test_standard_behavior(self):
        holder = dict()  # a real dict so we can control which object real ref
        def mk(k):
            o = O(k)
            holder[k] = o
            return o
        o = OrderedWeakValueDict((k, mk(k)) for k in xrange(10))
        self.assertEqual(len(o), 10)
        self.assertEqual(o[3].a, 3)
        o[3] = mk(-3)
        self.assertEqual(o[3].a, -3)
        del o[3]
        self.assertEqual(len(o), 9)
        o[100] = mk(100)
        self.assertEqual(len(o), 10)
        self.assertEqual(o.keys(), [0, 1, 2, 4, 5, 6, 7, 8, 9, 100])

        # remove a few items from the holder, they should be gone from o too
        del holder[1]
        del holder[5]
        del holder[6]

        self.assertEqual(o.keys(), [0, 2, 4, 7, 8, 9, 100])
        self.assertEqual([v.a for v in o.values()], [0, 2, 4, 7, 8, 9, 100])





#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from task import Task
from twisted.internet.defer import inlineCallbacks, TimeoutError, failure, returnValue
from twisted.internet import reactor
from pyvoltha.common.utils.asleep import asleep
from pyvoltha.adapters.extensions.omci.database.mib_db_dict import *
from pyvoltha.adapters.extensions.omci.omci_defs import AttributeAccess
from pyvoltha.adapters.extensions.omci.database.alarm_db_ext import AlarmDbExternal

AA = AttributeAccess


class AlarmCopyException(Exception):
    pass


class AlarmDownloadException(Exception):
    pass


class AlarmResyncException(Exception):
    pass


class AlarmResyncTask(Task):
    """
    OpenOMCI ALARM resynchronization Task

    This task should get a copy of the ALARM and compare compare it to a
    copy of the database. When the ALARM Upload command is sent to the ONU,
    it should make a copy and source the data requested from this database.
    The ONU can still source AVC's and the the OLT can still send config
    commands to the actual.
    """
    task_priority = Task.DEFAULT_PRIORITY
    name = "ALARM Resynchronization Task"

    max_retries = 3
    retry_delay = 7

    max_alarm_upload_next_retries = 3
    alarm_upload_next_delay = 10          # Max * delay < 60 seconds

    def __init__(self, omci_agent, device_id):
        """
        Class initialization

        :param omci_agent: (OmciAdapterAgent) OMCI Adapter agent
        :param device_id: (str) ONU Device ID
        """
        super(AlarmResyncTask, self).__init__(AlarmResyncTask.name,
                                              omci_agent,
                                              device_id,
                                              priority=AlarmResyncTask.task_priority,
                                              exclusive=False)
        self._local_deferred = None
        self._device = omci_agent.get_device(device_id)
        self._db_active = MibDbVolatileDict(omci_agent)
        self._db_active.start()

    def cancel_deferred(self):
        super(AlarmResyncTask, self).cancel_deferred()

        d, self._local_deferred = self._local_deferred, None
        try:
            if d is not None and not d.called:
                d.cancel()
        except:
            pass

    def start(self):
        """
        Start ALARM Re-Synchronization task
        """
        super(AlarmResyncTask, self).start()
        self._local_deferred = reactor.callLater(0, self.perform_alarm_resync)
        self._db_active.start()
        self._db_active.add(self.device_id)

    def stop(self):
        """
        Shutdown ALARM Re-Synchronization task
        """
        self.log.debug('stopping')

        self.cancel_deferred()
        self._device = None
        self._db_active.stop()
        self._db_active = None
        super(AlarmResyncTask, self).stop()

    @inlineCallbacks
    def perform_alarm_resync(self):
        """
        Perform the ALARM Resynchronization sequence

        The sequence to be performed is:
            - get a copy of the current ALARM database

            - perform ALARM upload commands to get ONU's database and save this
              to a local DB.
        During the alarm upload process, the maximum time between alarm upload next
        requests is 1 minute.
        """
        self.log.debug('perform-alarm-resync')

        try:
            self.strobe_watchdog()
            results = yield self.snapshot_alarm()
            olt_db_copy = results[0]
            number_of_commands = results[1]

            if olt_db_copy is None:
                e = AlarmCopyException('Failed to get local database copy')
                self.deferred.errback(failure.Failure(e))
            else:
                # Start the ALARM upload sequence, save alarms to the table
                self.strobe_watchdog()

                if number_of_commands > 0:
                    commands_retrieved = yield self.upload_alarm(number_of_commands)
                else:
                    commands_retrieved = 0

                if commands_retrieved != number_of_commands:
                    e = AlarmDownloadException('Only retrieved {} of {} instances'.
                                               format(commands_retrieved, number_of_commands))
                    self.deferred.errback(failure.Failure(e))
                else:
                    # Compare the databases
                    onu_db_copy = self._db_active.query(self.device_id)

                    on_olt_only, on_onu_only, attr_diffs = \
                        self.compare_mibs(olt_db_copy, onu_db_copy)

                    on_olt_only = on_olt_only if len(on_olt_only) else None
                    on_onu_only = on_onu_only if len(on_onu_only) else None
                    attr_diffs = attr_diffs if len(attr_diffs) else None

                    on_olt_only_diffs = on_olt_only if on_olt_only and len(on_olt_only) else None
                    on_onu_only_diffs = on_onu_only if on_onu_only and len(on_onu_only) else None
                    attr_diffs = attr_diffs if attr_diffs and len(attr_diffs) else None

                    if all(diff is None for diff in [on_olt_only_diffs, on_onu_only_diffs, attr_diffs]):
                        results = None
                    else:
                        results = {
                            'onu-only': on_onu_only_diffs,
                            'olt-only': on_olt_only_diffs,
                            'attr-diffs': attr_diffs,
                            'onu-db': onu_db_copy,
                            'olt-db': olt_db_copy
                        }
                    self.deferred.callback(results)

        except Exception as e:
            self.log.exception('resync', e=e)
            self.deferred.errback(failure.Failure(e))

    @inlineCallbacks
    def snapshot_alarm(self):
        """
        Snapshot the ALARM on the ONU and create a copy of our local ALARM database

        :return: (pair) (command_sequence_number)
        """
        olt_db_copy = None
        command_sequence_number = None

        try:
            max_tries = AlarmResyncTask.max_retries - 1

            for retries in xrange(0, max_tries + 1):
                # Send ALARM Upload so ONU snapshots its ALARM
                try:
                    command_sequence_number = yield self.send_alarm_upload()
                    self.strobe_watchdog()

                    if command_sequence_number is None:
                        if retries >= max_tries:
                            olt_db_copy = None
                            break

                except TimeoutError as e:
                    self.log.warn('timeout', e=e)
                    if retries >= max_tries:
                        raise

                    self.strobe_watchdog()
                    yield asleep(AlarmResyncTask.retry_delay)
                    continue

                # Get a snapshot of the local MIB database
                olt_db_copy = self._device.query_alarm_table()
                # if we made it this far, no need to keep trying
                break

        except Exception as e:
            self.log.exception('alarm-resync', e=e)
            raise

        # Handle initial failures

        if olt_db_copy is None or command_sequence_number is None:
            raise AlarmCopyException('Failed to snapshot ALARM copy after {} retries'.
                                     format(AlarmResyncTask.max_retries))

        returnValue((olt_db_copy, command_sequence_number))

    @inlineCallbacks
    def send_alarm_upload(self):
        """
        Perform ALARM upload command and get the number of entries to retrieve

        :return: (int) Number of commands to execute or None on error
        """
        ########################################
        # Begin ALARM Upload
        try:
            results = yield self._device.omci_cc.send_get_all_alarm()
            self.strobe_watchdog()
            command_sequence_number = results.fields['omci_message'].fields['number_of_commands']

            if command_sequence_number < 0:
                raise ValueError('Number of commands was {}'.format(command_sequence_number))

            returnValue(command_sequence_number)

        except TimeoutError as e:
            self.log.warn('alarm-resync-get-timeout', e=e)
            raise

    @inlineCallbacks
    def upload_alarm(self, command_sequence_number):
        ########################################
        # Begin ALARM Upload
        seq_no = None

        for seq_no in xrange(command_sequence_number):
            max_tries = AlarmResyncTask.max_alarm_upload_next_retries

            for retries in xrange(0, max_tries):
                try:
                    response = yield self._device.omci_cc.send_get_all_alarm_next(seq_no)
                    self.strobe_watchdog()

                    omci_msg = response.fields['omci_message'].fields
                    alarm_class_id = omci_msg['alarmed_entity_class']
                    alarm_entity_id = omci_msg['alarmed_entity_id']

                    alarm_bit_map = omci_msg['alarm_bit_map']
                    attributes = {AlarmDbExternal.ALARM_BITMAP_KEY: alarm_bit_map}

                    # Save to the database
                    self._db_active.set(self.device_id, alarm_class_id,
                                        alarm_entity_id, attributes)
                    break

                except TimeoutError:
                    self.log.warn('alarm-resync-timeout', seq_no=seq_no,
                                  command_sequence_number=command_sequence_number)

                    if retries < max_tries - 1:
                        yield asleep(AlarmResyncTask.alarm_upload_next_delay)
                        self.strobe_watchdog()
                    else:
                        raise

                except Exception as e:
                    self.log.exception('resync', e=e, seq_no=seq_no,
                                       command_sequence_number=command_sequence_number)

        returnValue(seq_no + 1)     # seq_no is zero based and alarm table.

    def compare_mibs(self, db_copy, db_active):
        """
        Compare the our db_copy with the ONU's active copy

        :param db_copy: (dict) OpenOMCI's copy of the database
        :param db_active: (dict) ONU's database snapshot
        :return: (dict), (dict), dict()  Differences
        """
        self.strobe_watchdog()

        # Class & Entities only in local copy (OpenOMCI)
        on_olt_only = self.get_lsh_only_dict(db_copy, db_active)

        # Class & Entities only on remote (ONU)
        on_onu_only = self.get_lsh_only_dict(db_active, db_copy)

        # Class & Entities on both local & remote, but one or more attributes
        # are different on the ONU.  This is the value that the local (OpenOMCI)
        # thinks should be on the remote (ONU)

        me_map = self.omci_agent.get_device(self.device_id).me_map
        attr_diffs = self.get_attribute_diffs(db_copy, db_active, me_map)

        return on_olt_only, on_onu_only, attr_diffs

    def get_lsh_only_dict(self, lhs, rhs):
        """
        Compare two MIB database dictionaries and return the ME Class ID and
        instances that are unique to the lhs dictionary. Both parameters
        should be in the common MIB Database output dictionary format that
        is returned by the mib 'query' command.

        :param lhs: (dict) Left-hand-side argument.
        :param rhs: (dict) Right-hand-side argument

        return: (list(int,int)) List of tuples where (class_id, inst_id)
        """
        results = list()

        for cls_id, cls_data in lhs.items():
            # Get unique classes
            #
            # Skip keys that are not class IDs
            if not isinstance(cls_id, int):
                continue

            if cls_id not in rhs:
                results.extend([(cls_id, inst_id) for inst_id in cls_data.keys()
                                if isinstance(inst_id, int)])
            else:
                # Get unique instances of a class
                lhs_cls = cls_data
                rhs_cls = rhs[cls_id]

                for inst_id, _ in lhs_cls.items():
                    # Skip keys that are not instance IDs
                    if isinstance(cls_id, int) and inst_id not in rhs_cls:
                        results.extend([(cls_id, inst_id)])

        return results

    def get_attribute_diffs(self, omci_copy, onu_copy, me_map):
        """
        Compare two OMCI MIBs and return the ME class and instance IDs that exists
        on both the local copy and the remote ONU that have different attribute
        values. Both parameters should be in the common MIB Database output
        dictionary format that is returned by the mib 'query' command.

        :param omci_copy: (dict) OpenOMCI copy (OLT-side) of the MIB Database
        :param onu_copy: (dict) active ONU latest copy its database
        :param me_map: (dict) ME Class ID MAP for this ONU

        return: (list(int,int,str)) List of tuples where (class_id, inst_id, attribute)
                                    points to the specific ME instance where attributes
                                    are different
        """
        results = list()

        # Get class ID's that are in both
        class_ids = {cls_id for cls_id, _ in omci_copy.items()
                     if isinstance(cls_id, int) and cls_id in onu_copy}

        for cls_id in class_ids:
            # Get unique instances of a class
            olt_cls = omci_copy[cls_id]
            onu_cls = onu_copy[cls_id]

            # Get set of common instance IDs
            inst_ids = {inst_id for inst_id, _ in olt_cls.items()
                        if isinstance(inst_id, int) and inst_id in onu_cls}

            for inst_id in inst_ids:
                omci_attributes = {k for k in olt_cls[inst_id][ATTRIBUTES_KEY].iterkeys()}
                onu_attributes = {k for k in onu_cls[inst_id][ATTRIBUTES_KEY].iterkeys()}

                # Get attributes that exist in one database, but not the other
                sym_diffs = (omci_attributes ^ onu_attributes)
                results.extend([(cls_id, inst_id, attr) for attr in sym_diffs])

                # Get common attributes with different values
                common_attributes = (omci_attributes & onu_attributes)
                results.extend([(cls_id, inst_id, attr) for attr in common_attributes
                               if olt_cls[inst_id][ATTRIBUTES_KEY][attr] !=
                                onu_cls[inst_id][ATTRIBUTES_KEY][attr]])
        return results



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

class ClosedTransactionError(Exception):
    pass


class ConfigTransaction(object):

    __slots__ = (
        '_proxy',
        '_txid'
    )

    def __init__(self, proxy, txid):
        self._proxy = proxy
        self._txid = txid

    def __del__(self):
        if self._txid:
            try:
                self.cancel()
            except:
                raise

    # ~~~~~~~~~~~~~~~~~~~~ CRUD ops within the transaction ~~~~~~~~~~~~~~~~~~~~

    def get(self, path='/', depth=None, deep=None):
        if self._txid is None:
            raise ClosedTransactionError()
        return self._proxy.get(path, depth=depth, deep=deep, txid=self._txid)

    def update(self, path, data, strict=False):
        if self._txid is None:
            raise ClosedTransactionError()
        return self._proxy.update(path, data, strict, self._txid)

    def add(self, path, data):
        if self._txid is None:
            raise ClosedTransactionError()
        return self._proxy.add(path, data, self._txid)

    def remove(self, path):
        if self._txid is None:
            raise ClosedTransactionError()
        return self._proxy.remove(path, self._txid)

    # ~~~~~~~~~~~~~~~~~~~~ transaction finalization ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    def cancel(self):
        """Explicitly cancel the transaction"""
        self._proxy.cancel_transaction(self._txid)
        self._txid = None

    def commit(self):
        """Commit all transaction changes"""
        try:
            self._proxy.commit_transaction(self._txid)
        finally:
            self._txid = None



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
A simple internal pub/sub event bus with topics and filter-based registration.
"""
import re

import structlog


log = structlog.get_logger()


class _Subscription(object):

    __slots__ = ('bus', 'predicate', 'callback', 'topic')
    def __init__(self, bus, predicate, callback, topic=None):
        self.bus = bus
        self.predicate = predicate
        self.callback = callback
        self.topic = topic


class EventBus(object):

    def __init__(self):
        self.subscriptions = {}  # topic -> list of _Subscription objects
                                 # topic None holds regexp based topic subs.
        self.subs_topic_map = {} # to aid fast lookup when unsubscribing

    def list_subscribers(self, topic=None):
        if topic is None:
            return sum(self.subscriptions.itervalues(), [])
        else:
            if topic in self.subscriptions:
                return self.subscriptions[topic]
            else:
                return []

    @staticmethod
    def _get_topic_key(topic):
        if isinstance(topic, str):
            return topic
        elif hasattr(topic, 'match'):
            return None
        else:
            raise AttributeError('topic not a string nor a compiled regex')

    def subscribe(self, topic, callback, predicate=None):
        """
        Subscribe to given topic with predicate and register the callback
        :param topic: String topic (explicit) or regexp based topic filter.
        :param callback: Callback method with signature def func(topic, msg)
        :param predicate: Optional method/function signature def predicate(msg)
        :return: Subscription object which can be used to unsubscribe
        """
        subscription = _Subscription(self, predicate, callback, topic)
        topic_key = self._get_topic_key(topic)
        self.subscriptions.setdefault(topic_key, []).append(subscription)
        self.subs_topic_map[subscription] = topic_key
        return subscription

    def unsubscribe(self, subscription):
        """
        Remove given subscription
        :param subscription: subscription object as was returned by subscribe
        :return: None
        """
        topic_key = self.subs_topic_map[subscription]
        self.subscriptions[topic_key].remove(subscription)

    def publish(self, topic, msg):
        """
        Publish given message to all subscribers registered with topic taking
        the predicate functions into account.
        :param topic: String topic
        :param msg: Arbitrary python data as message
        :return: None
        """
        from copy import copy

        def passes(msg, predicate):
            try:
                return predicate(msg)
            except Exception, e:
                return False  # failed predicate function treated as no match

        # lookup subscribers with explicit topic subscriptions
        subscribers = self.subscriptions.get(topic, [])

        # add matching regexp topic subscribers
        subscribers.extend(s for s in self.subscriptions.get(None, [])
                           if s.topic.match(topic))

        # iterate over a shallow-copy of subscribers
        for candidate in copy(subscribers):
            predicate = candidate.predicate
            if predicate is None or passes(msg, predicate):
                try:
                    candidate.callback(topic, msg)
                except Exception, e:
                    log.exception('callback-failed', e=repr(e), topic=topic)



default_bus = EventBus()


class EventBusClient(object):
    """
    Primary interface to the EventBus. Usage:

    Publish:
    >>> events = EventBusClient()
    >>> msg = dict(a=1, b='foo')
    >>> events.publish('a.topic', msg)

    Subscribe to get all messages on specific topic:
    >>> def got_event(topic, msg):
    >>>     print topic, ':', msg
    >>> events = EventBusClient()
    >>> events.subscribe('a.topic', got_event)

    Subscribe to get messages matching predicate on specific topic:
    >>> def got_event(topic, msg):
    >>>     print topic, ':', msg
    >>> events = EventBusClient()
    >>> events.subscribe('a.topic', got_event, lambda msg: msg.len() < 100)

    Use a DeferredQueue to buffer incoming messages
    >>> queue = DeferredQueue()
    >>> events = EventBusClient()
    >>> events.subscribe('a.topic', lambda _, msg: queue.put(msg))

    """
    def __init__(self, bus=None):
        """
        Obtain a client interface for the pub/sub event bus.
        :param bus: An optional specific event bus. Inteded for mainly test
        use. If not provided, the process default bus will be used, which is
        the preferred use (a process shall not need more than one bus).
        """
        self.bus = bus or default_bus

    def publish(self, topic, msg):
        """
        Publish given msg to given topic.
        :param topic: String topic
        :param msg: Arbitrary python data as message
        :return: None
        """
        self.bus.publish(topic, msg)

    def subscribe(self, topic, callback, predicate=None):
        """
        Subscribe to given topic with predicate and register the callback
        :param topic: String topic (explicit) or regexp based topic filter.
        :param callback: Callback method with signature def func(topic, msg)
        :param predicate: Optional method/function with signature
        def predicate(msg)
        :return: Subscription object which can be used to unsubscribe
        """
        return self.bus.subscribe(topic, callback, predicate)

    def unsubscribe(self, subscription):
        """
        Remove given subscription
        :param subscription: subscription object as was returned by subscribe
        :return: None
        """
        return self.bus.unsubscribe(subscription)

    def list_subscribers(self, topic=None):
        """
        Return list of subscribers. If topci is provided, it is filtered for
        those subscribing to the topic.
        :param topic: Optional topic
        :return: List of subscriptions
        """
        return self.bus.list_subscribers(topic)



# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from consul import Consul, ConsulException
from pyvoltha.common.utils.asleep import asleep
from requests import ConnectionError
from twisted.internet.defer import inlineCallbacks, returnValue

import etcd3
import structlog


class ConsulStore(object):
    """ Config kv store for consul with a cache for quicker subsequent reads

        TODO: This will block the reactor. Should either change
        whole call stack to yield or put the put/delete transactions into a
        queue to write later with twisted. Will need a transaction
        log to ensure we don't lose anything.
        Making the whole callstack yield is troublesome because other tasks can
        come in on the side and start modifying things which could be bad.
    """

    CONNECT_RETRY_INTERVAL_SEC = 1
    RETRY_BACKOFF = [0.05, 0.1, 0.2, 0.5, 1, 2, 5]

    def __init__(self, host, port, path_prefix):

        self.log = structlog.get_logger()
        self._consul = Consul(host=host, port=port)
        self.host = host
        self.port = port
        self._path_prefix = path_prefix
        self._cache = {}
        self.retries = 0

    def make_path(self, key):
        return '{}/{}'.format(self._path_prefix, key)

    def __getitem__(self, key):
        if key in self._cache:
            return self._cache[key]
        value = self._kv_get(self.make_path(key))
        if value is not None:
            # consul turns empty strings to None, so we do the reverse here
            self._cache[key] = value['Value'] or ''
            return value['Value'] or ''
        else:
            raise KeyError(key)

    def __contains__(self, key):
        if key in self._cache:
            return True
        value = self._kv_get(self.make_path(key))
        if value is not None:
            self._cache[key] = value['Value']
            return True
        else:
            return False

    def __setitem__(self, key, value):
        try:
            assert isinstance(value, basestring)
            self._cache[key] = value
            self._kv_put(self.make_path(key), value)
        except Exception, e:
            self.log.exception('cannot-set-item', e=e)

    def __delitem__(self, key):
        self._cache.pop(key, None)
        self._kv_delete(self.make_path(key))

    @inlineCallbacks
    def _backoff(self, msg):
        wait_time = self.RETRY_BACKOFF[min(self.retries,
                                           len(self.RETRY_BACKOFF) - 1)]
        self.retries += 1
        self.log.error(msg, retry_in=wait_time)
        yield asleep(wait_time)

    def _redo_consul_connection(self):
        self._consul = Consul(host=self.host, port=self.port)
        self._cache.clear()

    def _clear_backoff(self):
        if self.retries:
            self.log.info('reconnected-to-consul', after_retries=self.retries)
            self.retries = 0

    def _get_consul(self):
        return self._consul

    # Proxy methods for consul with retry support
    def _kv_get(self, *args, **kw):
        return self._retry('GET', *args, **kw)

    def _kv_put(self, *args, **kw):
        return self._retry('PUT', *args, **kw)

    def _kv_delete(self, *args, **kw):
        return self._retry('DELETE', *args, **kw)

    def _retry(self, operation, *args, **kw):
        while 1:
            try:
                consul = self._get_consul()
                self.log.debug('consul', consul=consul, operation=operation,
                         args=args)
                if operation == 'GET':
                    index, result = consul.kv.get(*args, **kw)
                elif operation == 'PUT':
                     result = consul.kv.put(*args, **kw)
                elif operation == 'DELETE':
                    result = consul.kv.delete(*args, **kw)
                else:
                    # Default case - consider operation as a function call
                    result = operation(*args, **kw)
                self._clear_backoff()
                break
            except ConsulException, e:
                self.log.exception('consul-not-up', e=e)
                self._backoff('consul-not-up')
            except ConnectionError, e:
                self.log.exception('cannot-connect-to-consul', e=e)
                self._backoff('cannot-connect-to-consul')
            except Exception, e:
                self.log.exception(e)
                self._backoff('unknown-error')
            self._redo_consul_connection()

        return result


class EtcdStore(object):
    """ Config kv store for etcd with a cache for quicker subsequent reads

        TODO: This will block the reactor. Should either change
        whole call stack to yield or put the put/delete transactions into a
        queue to write later with twisted. Will need a transaction
        log to ensure we don't lose anything.
        Making the whole callstack yield is troublesome because other tasks can
        come in on the side and start modifying things which could be bad.
    """

    CONNECT_RETRY_INTERVAL_SEC = 1
    RETRY_BACKOFF = [0.05, 0.1, 0.2, 0.5, 1, 2, 5]

    def __init__(self, host, port, path_prefix):

        self.log = structlog.get_logger()
        self._etcd = etcd3.client(host=host, port=port)
        self.host = host
        self.port = port
        self._path_prefix = path_prefix
        self._cache = {}
        self.retries = 0

    def make_path(self, key):
        return '{}/{}'.format(self._path_prefix, key)

    def __getitem__(self, key):
        if key in self._cache:
            return self._cache[key]
        (value, meta) = self._kv_get(self.make_path(key))
        if value is not None:
            self._cache[key] = value
            return value
        else:
            raise KeyError(key)

    def __contains__(self, key):
        if key in self._cache:
            return True
        (value, meta) = self._kv_get(self.make_path(key))
        if value is not None:
            self._cache[key] = value
            return True
        else:
            return False

    def __setitem__(self, key, value):
        try:
            assert isinstance(value, basestring)
            self._cache[key] = value
            self._kv_put(self.make_path(key), value)
        except Exception, e:
            self.log.exception('cannot-set-item', e=e)

    def __delitem__(self, key):
        self._cache.pop(key, None)
        self._kv_delete(self.make_path(key))

    @inlineCallbacks
    def _backoff(self, msg):
        wait_time = self.RETRY_BACKOFF[min(self.retries,
                                           len(self.RETRY_BACKOFF) - 1)]
        self.retries += 1
        self.log.error(msg, retry_in=wait_time)
        yield asleep(wait_time)

    def _redo_etcd_connection(self):
        self._etcd = etcd3.client(host=self.host, port=self.port)
        self._cache.clear()

    def _clear_backoff(self):
        if self.retries:
            self.log.info('reconnected-to-etcd', after_retries=self.retries)
            self.retries = 0

    def _get_etcd(self):
        return self._etcd

    # Proxy methods for etcd with retry support
    def _kv_get(self, *args, **kw):
        return self._retry('GET', *args, **kw)

    def _kv_put(self, *args, **kw):
        return self._retry('PUT', *args, **kw)

    def _kv_delete(self, *args, **kw):
        return self._retry('DELETE', *args, **kw)

    def _retry(self, operation, *args, **kw):

        # etcd data sometimes contains non-utf8 sequences, replace
        self.log.debug('backend-op',
                  operation=operation,
                  args=map(lambda x : unicode(x,'utf8','replace'), args),
                  kw=kw)

        while 1:
            try:
                etcd = self._get_etcd()
                self.log.debug('etcd', etcd=etcd, operation=operation,
                    args=map(lambda x : unicode(x,'utf8','replace'), args))
                if operation == 'GET':
                    (value, meta) = etcd.get(*args, **kw)
                    result = (value, meta)
                elif operation == 'PUT':
                    result = etcd.put(*args, **kw)
                elif operation == 'DELETE':
                    result = etcd.delete(*args, **kw)
                else:
                    # Default case - consider operation as a function call
                    result = operation(*args, **kw)
                self._clear_backoff()
                break
            except Exception, e:
                self.log.exception(e)
                self._backoff('unknown-error-with-etcd')
            self._redo_etcd_connection()

        return result


def load_backend(store_id, store_prefix, args):
    """ Return the kv store backend based on the command line arguments
    """

    def load_consul_store():
        instance_core_store_prefix = '{}/{}'.format(store_prefix, store_id)

        host, port = args.consul.split(':', 1)
        return ConsulStore(host, int(port), instance_core_store_prefix)

    def load_etcd_store():
        instance_core_store_prefix = '{}/{}'.format(store_prefix, store_id)

        host, port = args.etcd.split(':', 1)
        return EtcdStore(host, int(port), instance_core_store_prefix)

    loaders = {
        'none': lambda: None,
        'consul': load_consul_store,
        'etcd': load_etcd_store
    }

    return loaders[args.backend]()



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
OpenOMCI MIB Database API
"""

import structlog
from datetime import datetime

CREATED_KEY = 'created'
MODIFIED_KEY = 'modified'
MDS_KEY = 'mib_data_sync'
LAST_SYNC_KEY = 'last_mib_sync'
VERSION_KEY = 'version'
DEVICE_ID_KEY = 'device_id'
CLASS_ID_KEY = 'class_id'
INSTANCE_ID_KEY = 'instance_id'
ATTRIBUTES_KEY = 'attributes'
ME_KEY = 'managed_entities'
MSG_TYPE_KEY = 'message_types'


class DatabaseStateError(Exception):
    def __init__(self, *args):
        Exception.__init__(self, *args)


class MibDbApi(object):
    """
    MIB Database API Base Class

    Derive the ME MIB Database implementation from this API.  For an example
    implementation, look at the mib_db_dict.py implementation
    """
    def __init__(self, omci_agent):
        """
        Class initializer
        :param omci_agent: (OpenOMCIAgent) OpenOMCI Agent
        """
        self.log = structlog.get_logger()
        self._omci_agent = omci_agent
        self._started = False

        now = datetime.utcnow()
        self._created = now
        self._modified = now

    def start(self):
        """
        Start up/restore the database. For in-memory, will be a nop. For external
        DB, may need to create the DB and fetch create/modified values
        """
        if not self._started:
            self._started = True
        # For a derived class that is a persistent DB, Restore DB (connect,
        # get created/modified times, ....) or something along those lines.
        # Minimal restore could just be getting ONU device IDs' so they are cached
        # locally. Maximum restore would be a full in-memory version of database
        # for fast 'GET' request support.
        # Remember to restore the '_created' and '_modified' times (above) as well
        # from the database

    def stop(self):
        """
        Start up the database. For in-memory, will be a nop. For external
        DB, may need to create the DB and fetch create/modified values
        """
        if self._started:
            self._started = False

    @property
    def active(self):
        """
        Is the database active
        :return: (bool) True if active
        """
        return self._started

    @property
    def created(self):
        """
        Date (UTC) that the database was created
        :return: (datetime) creation date
        """
        return self._created

    @property
    def modified(self):
        """
        Date (UTC) that the database last added or removed a device
        or updated a device's ME information
        :return: (datetime) last modification date
        """
        return self._modified

    def add(self, device_id, overwrite=False):
        """
        Add a new ONU to database

        :param device_id: (str) Device ID of ONU to add
        :param overwrite: (bool) Overwrite existing entry if found.

        :raises KeyError: If device already exists and 'overwrite' is False
        """
        raise NotImplementedError('Implement this in your derive class')

    def remove(self, device_id):
        """
        Remove an ONU from the database

        :param device_id: (str) Device ID of ONU to remove from database
        """
        raise NotImplementedError('Implement this in your derive class')

    def set(self, device_id, class_id, entity_id, attributes):
        """
        Set/Create a database value.  This should only be called by the MIB synchronizer
        and its related tasks

        :param device_id: (str) ONU Device ID
        :param class_id: (int) ME Class ID
        :param entity_id: (int) ME Entity ID
        :param attributes: (dict) Attribute dictionary

        :returns: (bool) True if the value was saved to the database. False if the
                         value was identical to the current instance

        :raises KeyError: If device does not exist
        :raises DatabaseStateError: If the database is not enabled
        """
        raise NotImplementedError('Implement this in your derive class')

    def delete(self, device_id, class_id, entity_id):
        """
        Delete an entity from the database if it exists

        :param device_id: (str) ONU Device ID
        :param class_id: (int) ME Class ID
        :param entity_id: (int) ME Entity ID

        :returns: (bool) True if the instance was found and deleted. False
                         if it did not exist.

        :raises KeyError: If device does not exist
        :raises DatabaseStateError: If the database is not enabled
        """
        raise NotImplementedError('Implement this in your derive class')

    def query(self, device_id, class_id=None, instance_id=None, attributes=None):
        """
        Get database information.

        This method can be used to request information from the database to the detailed
        level requested

        :param device_id: (str) ONU Device ID
        :param class_id:  (int) Managed Entity class ID
        :param instance_id: (int) Managed Entity instance
        :param attributes: (list/set or str) Managed Entity instance's attributes

        :return: (dict) The value(s) requested. If class/inst/attribute is
                        not found, an empty dictionary is returned
        :raises KeyError: If the requested device does not exist
        :raises DatabaseStateError: If the database is not enabled
        """
        raise NotImplementedError('Implement this in your derive class')

    def on_mib_reset(self, device_id):
        """
        Reset/clear the database for a specific Device

        :param device_id: (str) ONU Device ID
        :raises DatabaseStateError: If the database is not enabled
        """
        # Your derived class should clear out all MIB data and update the
        # modified stats appropriately
        raise NotImplementedError('Implement this in your derive class')

    def save_mib_data_sync(self, device_id, value):
        """
        Save the MIB Data Sync to the database in an easy location to access

        :param device_id: (str) ONU Device ID
        :param value: (int) Value to save
        """
        raise NotImplementedError('Implement this in your derive class')

    def get_mib_data_sync(self, device_id):
        """
        Get the MIB Data Sync value last saved to the database for a device

        :param device_id: (str) ONU Device ID
        :return: (int) The Value or None if not found
        """
        raise NotImplementedError('Implement this in your derive class')

    def save_last_sync(self, device_id, value):
        """
        Save the Last Sync time to the database in an easy location to access

        :param device_id: (str) ONU Device ID
        :param value: (DateTime) Value to save
        """
        raise NotImplementedError('Implement this in your derive class')

    def get_last_sync(self, device_id):
        """
        Get the Last SYnc Time saved to the database for a device

        :param device_id: (str) ONU Device ID
        :return: (int) The Value or None if not found
        """
        raise NotImplementedError('Implement this in your derive class')

    def update_supported_managed_entities(self, device_id, managed_entities):
        """
        Update the supported OMCI Managed Entities for this device

        :param device_id: (str) ONU Device ID
        :param managed_entities: (set) Managed Entity class IDs
        """
        raise NotImplementedError('Implement this in your derive class')

    def update_supported_message_types(self, device_id, msg_types):
        """
        Update the supported OMCI Managed Entities for this device

        :param device_id: (str) ONU Device ID
        :param msg_types: (set) Message Type values (ints)
        """
        raise NotImplementedError('Implement this in your derive class')



#
# Copyright 2017 the original author or authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Utilities to handle gRPC server and client side code in a Twisted environment
"""
import structlog
from concurrent.futures import Future
from twisted.internet import reactor
from twisted.internet.defer import Deferred
from twisted.python.threadable import isInIOThread


log = structlog.get_logger()


def twisted_async(func):
    """
    This decorator can be used to implement a gRPC method on the twisted
    thread, allowing asynchronous programming in Twisted while serving
    a gRPC call.

    gRPC methods normally are called on the futures.ThreadPool threads,
    so these methods cannot directly use Twisted protocol constructs.
    If the implementation of the methods needs to touch Twisted, it is
    safer (or mandatory) to wrap the method with this decorator, which will
    call the inner method from the external thread and ensure that the
    result is passed back to the foreign thread.

    Example usage:

    When implementing a gRPC server, typical pattern is:

    class SpamService(SpamServicer):

        def GetBadSpam(self, request, context):
            '''this is called from a ThreadPoolExecutor thread'''
            # generally unsafe to make Twisted calls

        @twisted_async
        def GetSpamSafely(self, request, context):
            '''this method now is executed on the Twisted main thread
            # safe to call any Twisted protocol functions

        @twisted_async
        @inlineCallbacks
        def GetAsyncSpam(self, request, context):
            '''this generator can use inlineCallbacks Twisted style'''
            result = yield some_async_twisted_call(request)
            returnValue(result)

    """
    def in_thread_wrapper(*args, **kw):

        if isInIOThread():

            return func(*args, **kw)

        f = Future()

        def twisted_wrapper():
            try:
                d = func(*args, **kw)
                if isinstance(d, Deferred):

                    def _done(result):
                        f.set_result(result)
                        f.done()

                    def _error(e):
                        f.set_exception(e)
                        f.done()

                    d.addCallback(_done)
                    d.addErrback(_error)

                else:
                    f.set_result(d)
                    f.done()

            except Exception, e:
                f.set_exception(e)
                f.done()

        reactor.callFromThread(twisted_wrapper)
        try:
            result = f.result()
        except Exception, e:
            log.exception(e=e, func=func, args=args, kw=kw)
            raise

        return result

    return in_thread_wrapper


